type vid = string;;
type primname = string;;
type slevanot = Hi | Low;;
type pc = int;;
type 'a valexp = (*** value expression ***)
             Var of vid   (*** variable ***)
           | Unit          (*** unit ***)
           | SBool of bool * slevanot (*** boolean annotated with its security level ***)
           | SInt of int * slevanot   (*** integer annotated with its security level ***)
           | Bool of bool             (*** boolean ***)
           | Int of int               (*** integer ***)
           | Prim of primname * 'a option       (*** primitive ***)
           | App of 'a valexp * 'a valexp * pc (*** application ***)
           | Pair of 'a valexp * 'a valexp;;   (*** pair ***)

(**************************************************************************)
(************************** Process Expressions ***************************)
(**************************************************************************)
(*** Some processes are annotated with the position "pc" in the program ***)
(*** The positions are used in showing the result of analysis.          ***)
(*** More precisely, pc for each construct is defined as follows:       ***)
(***   Out: the position of "!"  ***)
(***   In: the position of "?"  ***)
(***   Nu: the start position of "new" ***)
(***   If: the start position of "if"  ***)
(***   Let: the start position of "let" ***)
(***   MatchL: currently, a match expression can only be translated from an if-expression,
 ***          so, the position of MatchL is that of the original if-expression puls one
 ***)
type cid = string;; (*** constructor id ***)

(*** syntax of processes ***)
(*** the parameter 'a denotes the type of type expressions for bound variables ***)
type 'a procexp = Zero
          | Out of vid * 'a valexp * 'a procexp * pc  
          | In of vid * vid * 'a * 'a procexp * pc
          | Par of 'a procexp * 'a procexp
          | Rep of 'a procexp
          | Nu of vid * 'a * 'a procexp * pc
          | If of 'a valexp * 'a procexp * 'a procexp * pc
          | MatchL of vid * 'a matchlist * pc (*** case analysis on data structures ***)
          | Let of vid * 'a * 'a valexp * 'a procexp * pc
          | Special of 'a procexp
and 'a matchlist = (cid * 'a procexp) list;;

module Vset = Set.Make(struct type t=string;; let compare = compare end);;

let make_vsubst = Utilities.make_subst;;
let rec apply_vsubst_v s v =
  match v with
   Var(x) -> (try List.assoc x s with Not_found -> v)
 | Unit -> v
 | SBool _ -> v
 | SInt _ -> v
 | Bool _ -> v
 | Int _ -> v
 | Prim (_,_) -> v
 | App(v1,v2,pc) -> App(apply_vsubst_v s v1, apply_vsubst_v s v2, pc)
 | Pair(v1,v2) -> Pair(apply_vsubst_v s v1, apply_vsubst_v s v2);;
              
let comp_vsubst s1 s2 = Utilities.comp_subst apply_vsubst_v s1 s2;;

let rec map_for_t f p =
 let g x = match x with Some(t) -> Some(f t) | None -> None in
 let rec map_for_t_in_v v =
   match v with
      Var vid -> Var vid
    | Unit -> Unit
    | SBool(b,s) -> SBool(b,s)
    | SInt(n,s) -> SInt(n,s)
    | Bool(b) -> Bool(b)
    | Int(n) -> Int(n)
    | Prim (s,t) -> Prim(s, g t)
    | App(v1,v2,pc) ->  App (map_for_t_in_v v1,map_for_t_in_v v2,pc) 
    | Pair(v1,v2) -> Pair (map_for_t_in_v v1,map_for_t_in_v v2)
 in
 let map_matchl_for_t f ml =
     List.map (fun (cid,p1) -> (cid, map_for_t f p1)) ml
 in
  match p with
    Zero -> Zero
  | Out(x,v,p1,pc) -> Out(x,map_for_t_in_v v, map_for_t f p1, pc)
  | In(x,y,t,p1,pc) -> In(x, y, f t, map_for_t f p1, pc)
  | Par(p1,p2) -> Par(map_for_t f p1, map_for_t f p2)
  | Rep(p1) -> Rep(map_for_t f p1)
  | Nu(x,t,p1,pc) -> Nu(x,f t, map_for_t f p1, pc)
  | If(v,p1,p2,pc) -> If(map_for_t_in_v v,map_for_t f p1, map_for_t f p2, pc)
  | MatchL(x, matchlist, pc) -> MatchL(x, map_matchl_for_t f matchlist, pc)
  | Let(x,t,v,p1,pc) -> Let(x, f t, map_for_t_in_v v, map_for_t f p1, pc)
  | Special(p) -> Special(map_for_t f p)

(*** normalize_procexp(p) normalizes p; 
 *** Currently, it only recognizes "if null(x) then ..." as a special form
 *** and converts it into a match expression
 *** In future, normalize_procexp should normalize pattern matches 
 ***)
let rec normalize_procexp p =
  match p with
     Zero -> Zero
   | Out(x, v, p1, pc) -> Out(x, v, normalize_procexp p1, pc)
   | In(x, y, t, p1, pc) -> In(x, y, t, normalize_procexp p1, pc)
   | Par(p1, p2) -> Par(normalize_procexp p1, normalize_procexp p2)
   | Rep(p1) -> Rep(normalize_procexp p1)
   | Nu(x, t, p1, pc) -> Nu(x, t, normalize_procexp p1, pc)
   | If(App(Prim ("null",_), Var x, pc1), p1, p2, pc2) -> 
        MatchL(x, [("nil", normalize_procexp(p1)); ("cons", normalize_procexp(p2))], pc2+1)
   | If(v, p1, p2, pc) -> 
        If(v, normalize_procexp p1, normalize_procexp p2, pc)
   | Let(x, t, v, p1, pc) ->
        Let(x, t, v, normalize_procexp p1, pc)
   | MatchL(x, ml, pc) -> p
   | Special(p) -> Special(normalize_procexp p);;

let extend_vsubst s x v = comp_vsubst (make_vsubst x v) s

let rename s x = try List.assoc x s with Not_found -> x
let newv x pc = x^"@"^(string_of_int pc);;

let rec alphaconv_procexp p s vs =
  match p with
     Zero -> Zero
   | Out(x, v, p1, pc) -> 
        Out(rename s x, apply_vsubst_v vs v, alphaconv_procexp p1 s vs, pc)
   | In(x, y, t, p1, pc) -> 
        let y' = newv y pc in
        let s' = (y,y')::s in
        let vs' = extend_vsubst vs y (Var(y')) in
           In(rename s x, y', t, alphaconv_procexp p1 s' vs', pc)
   | Par(p1, p2) -> 
        Par(alphaconv_procexp p1 s vs, alphaconv_procexp p2 s vs)
   | Rep(p1) -> Rep(alphaconv_procexp p1 s vs)
   | Nu(x, t, p1, pc) -> 
        let x' = newv x pc in 
        let s' = (x,x')::s in
        let vs' = extend_vsubst vs x (Var(x')) in
          Nu(x', t, alphaconv_procexp p1 s' vs', pc)
   | If(v, p1, p2, pc) -> 
        If(apply_vsubst_v vs v, alphaconv_procexp p1 s vs, alphaconv_procexp p2 s vs, pc)
   | Let(x, t, v, p1, pc) ->
        let x' = newv x pc in 
        let s' = (x,x')::s in
        let vs' = extend_vsubst vs x (Var(x')) in
           Let(x', t, apply_vsubst_v vs v, alphaconv_procexp p1 s' vs', pc)
   | MatchL(x, ml, pc) -> 
        let ml' = alphaconv_matchl ml s vs in
          MatchL(rename s x, ml', pc)
   | Special(p1) ->
        Special(alphaconv_procexp p1 s vs)
and alphaconv_matchl ml s vs =
  List.map (fun (cid, p) -> (cid, alphaconv_procexp p s vs)) ml;;

let alphaconv p = alphaconv_procexp p [] Utilities.subst_empty

let cleanv x _ = try String.sub x 0 (String.index x '@') with Not_found -> x

let rec cleanalpha_procexp p s vs =
  match p with
     Zero -> Zero
   | Out(x, v, p1, pc) -> 
        Out(rename s x, apply_vsubst_v vs v, cleanalpha_procexp p1 s vs, pc)
   | In(x, y, t, p1, pc) -> 
        let y' = cleanv y pc in
        let s' = (y,y')::s in
        let vs' = extend_vsubst vs y (Var(y')) in
           In(rename s x, y', t, cleanalpha_procexp p1 s' vs', pc)
   | Par(p1, p2) -> 
        Par(cleanalpha_procexp p1 s vs, cleanalpha_procexp p2 s vs)
   | Rep(p1) -> Rep(cleanalpha_procexp p1 s vs)
   | Nu(x, t, p1, pc) -> 
        let x' = cleanv x pc in 
        let s' = (x,x')::s in
        let vs' = extend_vsubst vs x (Var(x')) in
          Nu(x', t, cleanalpha_procexp p1 s' vs', pc)
   | If(v, p1, p2, pc) -> 
        If(apply_vsubst_v vs v, cleanalpha_procexp p1 s vs, cleanalpha_procexp p2 s vs, pc)
   | Let(x, t, v, p1, pc) ->
        let x' = cleanv x pc in 
        let s' = (x,x')::s in
        let vs' = extend_vsubst vs x (Var(x')) in
           Let(x', t, apply_vsubst_v vs v, cleanalpha_procexp p1 s' vs', pc)
   | MatchL(x, ml, pc) -> 
        let ml' = cleanalpha_matchl ml s vs in
          MatchL(rename s x, ml', pc)
   | Special(p1) -> 
          Special(cleanalpha_procexp p1 s vs)
and cleanalpha_matchl ml s vs =
  List.map (fun (cid, p) -> (cid, cleanalpha_procexp p s vs)) ml;;

let undo_alphaconv p = cleanalpha_procexp p [] Utilities.subst_empty

let infix_pr pr = List.mem pr ["+";"-";"*";"/";"=";"<";"<=";">";">="];;

let prec_output = 5;;
let prec_input = 5;;
let prec_rep = 4;;
let prec_if = 3;;
let prec_par = 2;;
let prec_let = 1;;
let prec_new = 1;;

let rec print_proc prec print_t p =
  match p with
    Zero -> print_string "O"
  | Out(x,v,p1,pc) ->
      (print_string x;
       print_string "!";
       print_value v;
       if p1=Zero then () else (print_string ".";print_proc prec_output print_t p1))
  | In(x,y,t,p1,pc) ->
      (print_string x;
       print_string "?";
       (match print_t with
           Some f -> print_string ("("^y^": "^(f t)^")")
         | None -> print_string y);
       if p1=Zero then () else (print_string ".";print_proc prec_input print_t p1))
  | Par(p1,p2) ->
      (if prec>prec_par then print_string "(";
       print_proc prec_par print_t p1;
       print_string " | ";
       print_proc prec_par print_t p2;
       if prec>prec_par then print_string ")")
  | Rep(p1) ->
      (print_string "*";
       if prec>prec_rep then print_string "(";
       print_proc prec_rep print_t p1;
       if prec>prec_rep then print_string ")")
  | Nu(x,t,p1,pc) ->
      (if prec>prec_new then (print_string "(") else ();
       (match print_t with
           Some f -> print_string ("new "^x^":"^(f t)^" in ")
          | None -> print_string ("new "^x^" in " ));
       print_proc prec_new print_t p1;
       if prec>prec_new then print_string ")")
  | If(v,p1,p2,pc) ->
      (if prec>prec_if then print_string "(";
       print_string "if ";
       print_value v;
       print_string " then ";
       print_proc prec_if print_t p1;
       print_string " else ";
       print_proc prec_if print_t p2;
       if prec>prec_if then print_string ")")
  | MatchL(x, [("nil", p1);("cons", p2)], pc) ->
(*** For now, match comes only from "if null(x) then ..." ***)
       print_proc prec print_t (If(App(Prim ("null",None), Var x, pc+2), p1, p2, pc-1))
  | Let(x,_,v,p1,pc) ->
      (if prec>prec_let then print_string "(";
       print_string ("let "^x^"=");
       print_value v;
       print_string " in ";
       print_proc prec_let print_t p1;
       if prec>prec_let then print_string ")")
  | Special(p1) ->
      (print_string "(#T";
       print_proc 0 print_t p1;
       print_string "#)")
      
  | _ -> print_string "<<match>>"
and print_value v =
  match v with
    Var(x) -> print_string x
  | Unit -> print_string "()"
  | SBool(b,sl) ->
      (if b then print_string "true" else print_string "false";
       if sl=Hi then print_string "/*H*/" else print_string "/*L*/")
  | Bool(b) ->       
      if b then print_string "true" else print_string "false"
  | SInt(n,sl) ->
      (print_string(string_of_int(n));
       if sl=Hi then print_string "/*H*/" else print_string "/*L*/")
  | Int(n) ->       
      print_string(string_of_int(n))
  | Prim(pr,_) -> print_string pr
  | App(v1,v2,pc) ->
      (match (v1,v2) with
         (Prim (pr,_), Pair(v21,v22)) -> 
            if infix_pr(pr) then
               (print_string "(";
                print_value v21;
                print_string pr;
                print_value v22;
                print_string ")")
            else
               (print_string "("; 
                print_value v1; 
                print_string " "; 
                print_value v2;
                print_string ")")
        | _ -> (print_string "("; 
                print_value v1; 
                print_string " "; 
                print_value v2;
                print_string ")")
        )
  | Pair(v1,v2) -> 
      (print_string "("; 
       print_value v1; 
       print_string ", "; 
       print_value v2;
       print_string ")")

let rec fvars_in_v v =
  match v with
    Var(x) -> Vset.singleton x
  | App(v1,v2,pc) -> Vset.union (fvars_in_v v1) (fvars_in_v v2)
  | Pair(v1,v2) ->Vset.union (fvars_in_v v1) (fvars_in_v v2)
  | _ -> Vset.empty

(*** fvars_in_proc(p) returns the set of free variables in process expression p ***)
let rec fvars_in_proc p =
  match p with
    Zero -> Vset.empty
  | Out(x,v,p1,pc) -> 
       Vset.union (Vset.add x (fvars_in_v v)) (fvars_in_proc p1)
  | In(x,y,_,p1,pc) ->
       Vset.add x (Vset.remove y (fvars_in_proc p1))
  | Par(p1,p2) ->
       Vset.union (fvars_in_proc p1) (fvars_in_proc p2)
  | Rep(p1) -> (fvars_in_proc p1)
  | Nu(x,_,p1,pc) ->
       Vset.remove x (fvars_in_proc p1)
  | If(v,p1,p2,pc) ->
       Vset.union (fvars_in_v v) (Vset.union (fvars_in_proc p1) (fvars_in_proc p2))
  | MatchL(x, ml, pc) ->
       Vset.add x (List.fold_right (fun (_, p) -> fun vs -> Vset.union (fvars_in_proc p) vs) ml Vset.empty)
  | Let(x,t,v,p1,pc) ->
       Vset.union (fvars_in_v v) (Vset.remove x (fvars_in_proc p1))
  | Special(p1) -> (fvars_in_proc p1)

let fv_proc p = Vset.elements (fvars_in_proc p)
       
let rec reduce_proc p =
  match p with
    Zero -> Zero
  | Out(x,v,p1,pc) ->
      let v'=reduce_value v in
      let p1'=reduce_proc p1 in
        Out(x,v',p1',pc)
  | In(x,y,t,p1,pc) ->
      let p1'=reduce_proc p1 in
        In(x,y,t,p1',pc)
  | Par(p1,p2) ->
      let p1'=reduce_proc p1 in
      let p2'=reduce_proc p2 in
        (match (p1',p2') with
           (Zero, _) -> p2'
         | (_, Zero) -> p1'
         | _ -> Par(p1',p2'))
  | Rep(p1) ->
      let p1'=reduce_proc p1 in
        if p1'=Zero then Zero else Rep(p1')
  | Nu(x,t,p1,pc) ->
      let p1'=reduce_proc p1 in
        if p1'=Zero then Zero else Nu(x,t,p1',pc)
  | If(v,p1,p2,pc) ->
      let v'=reduce_value v in
      let p1'=reduce_proc p1 in
      let p2'=reduce_proc p2 in
         if p1'=p2' then p1' else If(v',p1',p2',pc)
  | MatchL(x, ml, pc) ->
      MatchL(x, List.map (fun (con,p)->(con, reduce_proc p)) ml, pc)
  | Let(x,t,v,p1,pc) ->
      let v'=reduce_value v in
      let p1'=reduce_proc p1 in
        if Vset.mem x (fvars_in_proc p1') 
        then Let(x,t,v',p1',pc)
        else p1'
  | Special(p1) ->
      let p1'=reduce_proc p1 in
        if p1'=Zero then Zero else Special(p1')
and reduce_value v =
  match v with
     App(v1,v2,pc) ->
      let v1'=reduce_value v1 in
      let v2'=reduce_value v2 in
       (match (v1',v2') with
            (Prim ("fst",_), Pair(v21,v22)) -> v21
          | (Prim ("snd",_), Pair(v21,v22)) -> v22
          | _ -> App(v1',v2',pc)
        )
   | Pair(v1,v2) ->
      let v1'=reduce_value v1 in
      let v2'=reduce_value v2 in
         Pair(v1', v2')
   | _ -> v;;

let pp_newline indent = print_string ("\n"^indent);;
let inc_indent indent = indent^"  ";;

let rec sizeof p =
  match p with
     Zero -> 0
   | Out(x,v,p1,pc) -> 1+sizeof p1
   | In(x,y,t,p1,pc) ->  1+sizeof p1
   | Par(p1,p2) -> 1+(sizeof p1)+(sizeof p2)
   | Rep (p1) -> 1+sizeof p1
   | Nu(x,t,p1,pc) -> 2+sizeof p1
   | If(v,p1,p2,pc) -> 3+(sizeof p1)+(sizeof p2)
   | MatchL(x,ml,pc) -> (List.fold_right (fun (con,p)->fun n->n+5+(sizeof p)) ml 3)
   | Let(x,t,v,p1,pc) -> 2+sizeof p1
   | Special(p1) -> 1 + (sizeof p1)
(*** pretty printing of processes and values ***)
let rec pprint_proc (indent:string) (prec:int) (size_limit: int) print_t p =
  if sizeof(p)<size_limit then
     print_proc prec print_t p
  else
    match p with
      Zero -> print_string "O"
    | Out(x,v,p1,pc) ->
        (print_string x;
         print_string "!";
         print_value v;
         if p1=Zero then () 
         else 
           (print_string ".";
              let indent' = inc_indent(indent) in
                (pp_newline indent'; 
                 pprint_proc indent' prec_output size_limit print_t p1)))
    | In(x,y,t,p1,pc) ->
        (print_string x;
         print_string "?";
         (match print_t with
           Some f -> print_string ("("^y^": "^(f t)^")")
          | None -> print_string y);
         if p1=Zero then () 
         else 
           (print_string ".";
              let indent' = inc_indent(indent) in
                (pp_newline indent'; 
                 pprint_proc indent' prec_input size_limit print_t p1)))
    | Par(p1,p2) ->
        (if prec>prec_par then print_string "(";
         pprint_proc indent prec_par size_limit print_t p1;
         print_string (" |");
         pp_newline (indent);
         pprint_proc indent prec_par size_limit print_t p2;
         if prec>prec_par then print_string ")")
    | Rep(p1) ->
        (print_string "*";
         if prec>prec_rep then print_string "(";
         pprint_proc (inc_indent indent) prec_rep size_limit print_t p1;
         if prec>prec_rep then print_string ")")
    | Nu(x,t,p1,pc) ->
        (if prec>prec_new then print_string "(";
         (match print_t with
           Some f -> print_string ("new "^x^":"^(f t)^" in ")
          | None -> print_string ("new "^x^" in " ));
         let indent' = inc_indent(indent) in
           (pp_newline indent';
            pprint_proc indent' prec_new size_limit print_t p1;
            if prec>prec_new then print_string ")"))
    | If(v,p1,p2,pc) ->
        let indent' = inc_indent(indent) in 
          (if prec>prec_if then print_string "(";
           print_string "if ";
           print_value v;
           print_string " then ";
           pp_newline(indent');
           pprint_proc indent' prec_if size_limit print_t p1;
           pp_newline(indent);
           print_string "else ";
           pp_newline(indent');
           pprint_proc indent' prec_if size_limit print_t p2;
          if prec>prec_if then print_string ")")
  | MatchL(x, [("nil", p1);("cons", p2)], pc) ->
(*** For now, match comes only from "if null(x) then ..." ***)
       pprint_proc indent prec size_limit print_t
                  (If(App(Prim ("null",None), Var x, pc+2), p1, p2, pc-1))
    | Let(x,t,v,p1,pc) ->
        let indent' = inc_indent(indent) in 
          (if prec>prec_let then print_string "(";
           print_string ("let "^x^"=");
           print_value v;
           print_string " in ";
           pp_newline(indent');
           pprint_proc indent' prec_let size_limit print_t p1;
          if prec>prec_let then print_string ")")
    | Special(p1) ->
        (print_string "(#T";
         pprint_proc (inc_indent indent) 0 size_limit print_t p1;
         print_string "#)")
    | _ -> print_string "<<match>>"

let pp_proc print_t p = pprint_proc "" 0 7 print_t p;;

let rec get_subproc p pc =
  match p with
    Zero -> raise Not_found
  | Out(_,_,p1,pc1) -> if pc=pc1 then p else get_subproc p1 pc
  | In(_,_,_,p1,pc1) -> if pc=pc1 then p else get_subproc p1 pc
  | Par(p1,p2) -> (try get_subproc p1 pc with Not_found -> get_subproc p2 pc)
  | Rep(p1) -> get_subproc p1 pc 
  | Nu(_,_,p1,pc1) -> if pc=pc1 then p else get_subproc p1 pc
  | If(_,p1,p2,pc1) ->
        if pc=pc1 then p else (try get_subproc p1 pc with Not_found -> get_subproc p2 pc)
  | MatchL(_,ml,pc1) -> 
        if pc=pc1 then p else get_subproc_in_ml ml pc
  | Let(_,_,_,p1,pc1) -> if pc=pc1 then p else get_subproc p1 pc
  | Special(p1) -> get_subproc p1 pc
and get_subproc_in_ml ml pc =
  match ml with
   [] -> raise Not_found
 | (_,p1)::ml' -> (try get_subproc p1 pc with Not_found -> get_subproc_in_ml ml' pc)

let rec rep_restricted p =
  match p with
    Zero -> true
  | Out(_,_,p1,pc1) -> rep_restricted p1
  | In(_,_,_,p1,pc1) -> rep_restricted p1
  | Par(p1,p2) -> rep_restricted p1 && rep_restricted p2
  | Rep(In(_,_,_,p1,_)) -> rep_restricted p1
  | Rep(_) -> false
  | Nu(_,_,p1,pc1) -> rep_restricted p1
  | If(_,p1,p2,pc1) ->rep_restricted p1 && rep_restricted p2
  | MatchL(_,ml,pc1) -> 
        List.for_all (fun (_,p1)-> rep_restricted p1) ml
  | Let(_,_,_,p1,pc1) -> rep_restricted p1
  | Special(p1) -> rep_restricted p1


let rec add_annotation p pc =
  match p with
     Zero -> Zero
   | Out(x, v, p1, pc1) -> 
        if pc=pc1 then 
            Special(p)
        else
            Out(x, v, add_annotation p1 pc, pc1)
   | In(x, y, t, p1, pc1) -> 
        if pc=pc1 then 
            Special(p)
        else
            In(x, y, t, add_annotation p1 pc, pc1)
   | Par(p1, p2) -> 
        Par(add_annotation p1 pc, add_annotation p2 pc)
   | Rep(p1) -> Rep(add_annotation p1 pc)
   | Nu(x, t, p1, pc1) -> 
        if pc=pc1 then 
            Special(p)
        else
          Nu(x, t, add_annotation p1 pc, pc1)
   | If(v, p1, p2, pc1) -> 
        if pc=pc1 
        then Special(p)
        else If(v, add_annotation p1 pc, add_annotation p2 pc, pc1)
   | Let(x, t, v, p1, pc1) ->
        if pc=pc1 
        then Special(p)
        else   Let(x, t, v, add_annotation p1 pc, pc1)
   | MatchL(x, ml, pc1) -> 
        if pc=pc1 
        then Special(p)
        else MatchL(x, List.map (fun (cid,p)-> (cid,add_annotation p pc)) ml, pc1)
   | Special(p1) ->
        Special(add_annotation p1 pc)

let rec pcs p =
  match p with
     Zero -> []
   | Out(x, v, p1, pc1) -> 
        Utilities.merge_and_unify compare [pc1] (pcs p1)
   | In(x, y, t, p1, pc1) -> 
        Utilities.merge_and_unify compare [pc1] (pcs p1)
   | Par(p1, p2) -> 
        Utilities.merge_and_unify compare (pcs p2) (pcs p1)
   | Rep(p1) -> 
        (pcs p1)
   | Nu(x, t, p1, pc1) -> 
        Utilities.merge_and_unify compare [pc1] (pcs p1)
   | If(v, p1, p2, pc1) -> 
        Utilities.merge_and_unify compare [pc1] 
            (Utilities.merge_and_unify compare (pcs p2) (pcs p1))
   | Let(x, t, v, p1, pc1) ->
        Utilities.merge_and_unify compare [pc1] (pcs p1)
   | MatchL(x, ml, pc1) -> 
        Utilities.merge_and_unify compare [pc1] (pcs_in_ml ml)
   | Special(p1) ->
        (pcs p1)
and pcs_in_ml ml =
  match ml with
    [] -> []
  | (_,p1)::ml' -> Utilities.merge_and_unify compare (pcs p1) (pcs_in_ml ml')

